/*
  CPP_CONTEST=2013Cal-2012
  CPP_PROBLEM=J
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 6
*/

/* RECORD
David López Fernández
student of Methodology of  Parallel Programming,
at the University of Murcia, Spain
February 1, 2016
time 71 msec
speed-up 294.37
*/ 

/*
The MPI version of this problem has been solved by sending to each process its own proportion of
data which is going to work with. On the basis of MPI implementation, the hybrid version consists of each process
starting threads. 
The sequential improvement deals with placing the pin in the middle.
 */

#include <stdlib.h>
#include <omp.h>
#include <mpi.h>

extern void escribir();

void mezclar(int *a, int t, int trozo) {
  int i, j, k;

  int *b = (int *) malloc(sizeof(int) * (t + trozo) );

  i = 0;
  j = t;
  k = 0;

  while (i < t && j < t + trozo ) {
    if (a[i] < a[j]) {
      b[k] = a[i];
      i++;
    } else {
      b[k] = a[j];
      j++;
    }
    k++;
  }
  for (; i < t; i++) {
    b[k] = a[i];
    k++;
  }
  for (; j < t + trozo; j++) {
    b[k] = a[j];
    k++;
  }
  for (i = 0; i < t + trozo; i++)
    a[i] = b[i];


  free(b);
}

void distribuir(int *a, int t, int nodo, int np) {

  int tam = (t / np) * t;
  // DISTRIBUYE LOS DATOS QUE LE CORRESPONE A CADA UNO DE LOS NODOS
  MPI_Scatter(a, tam, MPI_INT, a, tam, MPI_INT, 0, MPI_COMM_WORLD);
}

void quicksort(int *a, int lo, int hi) {
  int i = lo, j = hi, h;
  int pivote = a[(lo + hi) / 2];

  do {
    while (a[i] < pivote) i++;
    while (a[j] > pivote) j--;
    if (i <= j) {
      h = a[i];
      a[i] = a[j];
      a[j] = h;
      i++; j--;
    }
  } while (i <= j);

  //   #pragma omp parallel sections
  // {
  //    #pragma omp section
  if (lo < j) quicksort(a, lo, j);
  //   #pragma omp section
  if (i < hi) quicksort(a, i, hi);
  //}


}
//t es el tamaÃƒÂ±o del array, np el nÃƒÂºmero de procesos, y nodo el identificador de proceso
void combinar(int *a, int t, int nodo, int np) {

  int tam = (t / np) * t; //elementos de cada proceso
  int desp = (t / np) * t; //numero elementos a combinar en prox interacion
  int resto = (t % np) * t; //filas restantes que se mezclan al final

  // RECIBE LOS DATOS DE CADA UNO DE LOS NODOS
  MPI_Gather(a, tam, MPI_INT, a, tam, MPI_INT, 0, MPI_COMM_WORLD);

  if (nodo == 0 ) {
    quicksort(&a[np * (t / np) * t], 0, resto - 1); //ordenaciÃƒÂ³n del resto filas sobrantes

    while (tam < (t * t) - resto) {
      mezclar(a, tam, desp);
      tam += desp;
    }

    mezclar(a, tam, resto); //mezcla de filas sobrantes
  }
}
void transpose(int n, int *m, int *r)
{
  int i, j;

  for (i = 0; i < n; i++)
  {
    for (j = 0; j < n; j++)
    {
      r[i * n + j] = m[j * n + i];
    }
  }
}

void mmt(int t, int *a, int *r, int np, int nodo, int resto)
{
  int i, j, k, s, *at;
  at = (int *) malloc(sizeof(int) * t * t);

  transpose(t, a, at);
  int  filaInicial = (nodo * (t / np));
  int indice = 0;

#ifdef DEBUG
  // While debugging the transpose matrix is written
  escribir(at, t);
#endif

  for (i = filaInicial; i < filaInicial + (t / np); i++)
  {

    for (j = 0; j < t; j++)
    {
      s = 0.;
      for (k = 0; k < t ; k++) {
        s = s + a[i * t + k] * at[j * t + k];
      }
      r[indice * t + j] = s;
    }
    indice++;
  }
  int ultimasFilas = t - 1;

  //calculo de las filas restantes. Se empieza por la ultima fila. Si resto = 2
  //se calculan las dos ultimas filas. La realiza el nodo 0, ya que este conoce
  //toda la matriz a
  if (nodo == 0) {

    for (i = 0 ; i < resto ; i++) {
      for (j = 0 ; j < t ; j++) {
        s = 0.;
        for (k = 0; k < t ; k++)
          s = s + a[ultimasFilas * t + k] * at[ j * t + k];

        r[ultimasFilas * t + j] = s;

      }
      ultimasFilas--; //fila matriz a anterior
    }

  }

  free(at);
}

void sec(int steps, int t, int *a, int nodo, int np)
{
  int i, *mtemp, *p, numeroPasadas;

  MPI_Bcast(&t, 1, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Bcast(&steps, 1, MPI_INT, 0, MPI_COMM_WORLD);

  int resto = t % np;

  if (nodo != 0) {
    a = (int *) malloc(t / np * t * (sizeof(int)));
    mtemp = (int *) malloc(t / np * t * (sizeof(int)));
  }
  else
    mtemp = (int *) malloc(sizeof(int) * t * t);


  for (numeroPasadas = 0; numeroPasadas < steps; numeroPasadas++) {

    distribuir(a, t, nodo, np);

    quicksort(a, 0, ((t / np)*t) - 1); //ordenaciÃƒÂ³n

    combinar(a, t, nodo, np);

    //Aqui esta todo ordenado

    if (nodo != 0) 
        a = (int *) malloc(t * t * (sizeof(int))); //importante cambiar tamaÃ±o de a
  
    //es necesario compartir a para la multiplicacion
    MPI_Bcast(a, t * t, MPI_INT, 0, MPI_COMM_WORLD);

    mmt(t, a, mtemp, np, nodo, resto); //mult

    //nodo 0 recibe los resultados de cada nodo en la zona de memoria correcta
    //Se podrÃ­a haber usado gather pero por cambiar lo he hecho asÃ­
    if (nodo == 0) {
      for (i = 1; i < np; i++) {
        MPI_Recv(&mtemp[i * (t / np) * t], t / np * t, MPI_INT, i, 40, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      }

      p = mtemp;
      mtemp = a;
      a = p;

    }
    else
      MPI_Send(mtemp, t / np * t, MPI_INT, 0, 40, MPI_COMM_WORLD); //cada uno envia la soluciÃƒÂ³n

#ifdef DEBUG
    // While debugging the matrix and the numbers are written
    escribir(a, t);
#endif
  }

  if (nodo == 0) {
    if (steps % 2 == 1)
    {
      p = mtemp;
      mtemp = a;
      a = p;
    }

  }
  else {
    free(mtemp);
    free(a);
  }
}
